//+build wasm

// Package JS provides additional functionality on top of syscall/js package for WASM.
package js

import (
	"sync"
	"syscall/js"
)

var (
	global    = js.Global()
	null      = js.Null()
	undefined = js.Undefined()
)

var (
	mu      sync.RWMutex
	classes = make(map[string]Value)
)

// Get is a shorthand for Global().Get().
func Get(name string) Value {
	return Value{global.Get(name)}
}

// Set is a shorthand for Global().Set().
func Set(name string, v Value) {
	global.Set(name, v.JSRef())
}

// Class searches for a class in global scope.
// It caches results, so the lookup should be faster than calling Get.
func Class(class string) Value {
	mu.RLock()
	v := classes[class]
	mu.RUnlock()
	if v.isZero() {
		v = Get(class)
		mu.Lock()
		classes[class] = v
		mu.Unlock()
	}
	return v
}

// New searches for a class in global scope and creates a new instance of that class.
func New(class string, args ...interface{}) Value {
	v := Class(class)
	return v.New(args...)
}

// Ref is an alias for syscall/js.Value.
type Ref = js.Value

// JSRef is a common interface for object that are backed by a JS object.
type JSRef interface {
	// JSRef returns a JS object reference as defined by syscall/js.
	JSRef() Ref
}

func toJS(o interface{}) interface{} {
	switch v := o.(type) {
	case JSRef:
		o = v.JSRef()
	case []Value:
		refs := make([]Ref, 0, len(v))
		for _, ref := range v {
			refs = append(refs, ref.JSRef())
		}
		o = refs
	}
	return o
}

var _ JSRef = Value{}

// Value is a convenience wrapper for syscall/js.Value.
// It provides some additional functionality, while storing no additional state.
// Its safe to instantiate Value directly, by wrapping syscall/js.Value.
type Value struct {
	Ref
}

func (v Value) isZero() bool {
	return v == (Value{})
}

// JSRef returns a JS object reference as defined by syscall/js.
func (v Value) JSRef() Ref {
	return v.Ref
}

// String converts a value to a string.
func (v Value) String() string {
	if !v.Valid() {
		return ""
	}
	return v.Ref.String()
}

// IsNull checks if a value represents JS null object.
func (v Value) IsNull() bool {
	return v.Ref == null
}

// IsUndefined checks if a value represents JS undefined object.
func (v Value) IsUndefined() bool {
	return v.Ref == undefined
}

// Valid checks if object is defined and not null.
func (v Value) Valid() bool {
	return !v.isZero() && !v.IsNull() && !v.IsUndefined()
}

// Get returns the JS property by name.
func (v Value) Get(name string) Value {
	return Value{v.Ref.Get(name)}
}

// Set sets the JS property to ValueOf(x).
func (v Value) Set(name string, val interface{}) {
	v.Ref.Set(name, toJS(ValueOf(val)))
}

// TODO: Del

// Index returns JS index i of value v.
func (v Value) Index(i int) Value {
	return Value{v.Ref.Index(i)}
}

// SetIndex sets the JavaScript index i of value v to ValueOf(x).
func (v Value) SetIndex(i int, val interface{}) {
	v.Ref.SetIndex(i, toJS(ValueOf(val)))
}

// Call does a JavaScript call to the method m of value v with the given arguments.
// It panics if v has no method m.
// The arguments get mapped to JavaScript values according to the ValueOf function.
func (v Value) Call(name string, args ...interface{}) Value {
	for i, a := range args {
		args[i] = toJS(ValueOf(a))
	}
	return Value{v.Ref.Call(name, args...)}
}

// Invoke does a JavaScript call of the value v with the given arguments.
// It panics if v is not a function.
// The arguments get mapped to JavaScript values according to the ValueOf function.
func (v Value) Invoke(args ...interface{}) Value {
	for i, a := range args {
		args[i] = toJS(ValueOf(a))
	}
	return Value{v.Ref.Invoke(args...)}
}

// New uses JavaScript's "new" operator with value v as constructor and the given arguments.
// It panics if v is not a function.
// The arguments get mapped to JavaScript values according to the ValueOf function.
func (v Value) New(args ...interface{}) Value {
	for i, a := range args {
		args[i] = toJS(ValueOf(a))
	}
	return Value{v.Ref.New(args...)}
}

// InstanceOf reports whether v is an instance of type t according to JavaScript's instanceof operator.
func (v Value) InstanceOf(class Value) bool {
	return v.Ref.InstanceOf(class.Ref)
}

// InstanceOfClass reports whether v is an instance of named type according to JavaScript's instanceof operator.
func (v Value) InstanceOfClass(class string) bool {
	return v.InstanceOf(Class(class))
}

// Slice converts JS Array to a Go slice of JS values.
func (v Value) Slice() []Value {
	if !v.Valid() {
		return nil
	}
	n := v.Length()
	vals := make([]Value, 0, n)
	for i := 0; i < n; i++ {
		vals = append(vals, v.Index(i))
	}
	return vals
}

// ValueOf returns x as a JavaScript value:
//
//  | Go                     | JavaScript             |
//  | ---------------------- | ---------------------- |
//  | js.Value               | [its value]            |
//  | js.TypedArray          | typed array            |
//  | js.Callback            | function               |
//  | nil                    | null                   |
//  | bool                   | boolean                |
//  | integers and floats    | number                 |
//  | string                 | string                 |
//  | []interface{}          | new array              |
//  | map[string]interface{} | new object             |
func ValueOf(o interface{}) Value {
	return Value{js.ValueOf(toJS(o))}
}
